@@ -1,17 +1,17 @@
 #![doc = include_str!("../README.md")]
 
+use std::fmt::{self, Error, Formatter};
+
 use borsh::{BorshDeserialize, BorshSerialize};
 use near_account_id::AccountId;
 use near_rpc_error_macro::RpcError;
 use serde::{Deserialize, Serialize};
-use std::any::Any;
-use std::fmt::{self, Error, Formatter};
 
-#[derive(Debug, PartialEq, Eq)]
+#[derive(Debug, Clone, PartialEq, Eq)]
 pub enum VMError {
     FunctionCallError(FunctionCallError),
-    /// Type erased error from `External` trait implementation.
-    ExternalError(AnyError),
+    /// Serialized external error from External trait implementation.
+    ExternalError(Vec<u8>),
     /// An error that is caused by an operation on an inconsistent state.
     /// E.g. an integer overflow by using a value from the given context.
     InconsistentStateError(InconsistentStateError),
@@ -19,7 +19,7 @@ pub enum VMError {
     CacheError(CacheError),
 }
 
-#[derive(Debug, PartialEq, Eq)]
+#[derive(Debug, Clone, PartialEq, Eq)]
 pub enum FunctionCallError {
     /// Wasm compilation error
     CompilationError(CompilationError),
@@ -46,7 +46,6 @@ pub enum FunctionCallError {
 /// add new variants at the end (but do that very carefully). This type must be never used
 /// directly, and must be converted to `ContractCallError` instead using `into()` converter.
 /// It describes stable serialization format, and only used by serialization logic.
-#[cfg_attr(feature = "deepsize_feature", derive(deepsize::DeepSizeOf))]
 #[derive(Debug, Clone, PartialEq, Eq, BorshDeserialize, BorshSerialize, Serialize, Deserialize)]
 pub enum FunctionCallErrorSer {
     /// Wasm compilation error
@@ -67,7 +66,7 @@ pub enum FunctionCallErrorSer {
     ExecutionError(String),
 }
 
-#[derive(Debug, PartialEq, Eq)]
+#[derive(Debug, PartialEq, Eq, Clone, Copy)]
 pub enum CacheError {
     ReadError,
     WriteError,
@@ -75,7 +74,6 @@ pub enum CacheError {
     SerializationError { hash: [u8; 32] },
 }
 /// A kind of a trap happened during execution of a binary
-#[cfg_attr(feature = "deepsize_feature", derive(deepsize::DeepSizeOf))]
 #[derive(
     Debug, Clone, PartialEq, Eq, BorshDeserialize, BorshSerialize, Deserialize, Serialize, RpcError,
 )]
@@ -100,7 +98,6 @@ pub enum WasmTrap {
     GenericTrap,
 }
 
-#[cfg_attr(feature = "deepsize_feature", derive(deepsize::DeepSizeOf))]
 #[derive(
     Debug, Clone, PartialEq, Eq, BorshDeserialize, BorshSerialize, Deserialize, Serialize, RpcError,
 )]
@@ -110,7 +107,6 @@ pub enum MethodResolveError {
     MethodInvalidSignature,
 }
 
-#[cfg_attr(feature = "deepsize_feature", derive(deepsize::DeepSizeOf))]
 #[derive(
     Debug, Clone, PartialEq, Eq, BorshDeserialize, BorshSerialize, Deserialize, Serialize, RpcError,
 )]
@@ -121,7 +117,6 @@ pub enum CompilationError {
     UnsupportedCompiler { msg: String },
 }
 
-#[cfg_attr(feature = "deepsize_feature", derive(deepsize::DeepSizeOf))]
 #[derive(
     Debug, Clone, PartialEq, Eq, BorshDeserialize, BorshSerialize, Deserialize, Serialize, RpcError,
 )]
@@ -149,10 +144,10 @@ pub enum PrepareError {
     /// Error creating memory.
     Memory,
     /// Contract contains too many functions.
+    #[cfg(feature = "protocol_feature_limit_contract_functions_number")]
     TooManyFunctions,
 }
 
-#[cfg_attr(feature = "deepsize_feature", derive(deepsize::DeepSizeOf))]
 #[derive(
     Debug, Clone, PartialEq, Eq, BorshDeserialize, BorshSerialize, Deserialize, Serialize, RpcError,
 )]
@@ -227,22 +222,23 @@ pub enum HostError {
     AltBn128SerializationError { msg: String },
 }
 
-#[derive(Debug, PartialEq)]
+#[derive(Debug, Clone, PartialEq, BorshDeserialize, BorshSerialize, Deserialize, Serialize)]
 pub enum VMLogicError {
     /// Errors coming from native Wasm VM.
     HostError(HostError),
-    /// Type erased error from `External` trait implementation.
-    ExternalError(AnyError),
+    /// Serialized external error from External trait implementation.
+    ExternalError(Vec<u8>),
     /// An error that is caused by an operation on an inconsistent state.
     InconsistentStateError(InconsistentStateError),
 }
 
 impl std::error::Error for VMLogicError {}
 
-/// An error that is caused by an operation on an inconsistent state, such as
-/// integer overflow.
-#[derive(Debug, Clone, PartialEq, Eq)]
+/// An error that is caused by an operation on an inconsistent state.
+/// E.g. a deserialization error or an integer overflow.
+#[derive(Debug, Clone, PartialEq, Eq, BorshDeserialize, BorshSerialize, Deserialize, Serialize)]
 pub enum InconsistentStateError {
+    StorageError(String),
     /// Math operation with a value from the state resulted in a integer overflow.
     IntegerOverflow,
 }
@@ -267,14 +263,14 @@ impl From<PrepareError> for VMError {
     }
 }
 
-impl From<VMLogicError> for VMError {
-    fn from(err: VMLogicError) -> Self {
+impl From<&VMLogicError> for VMError {
+    fn from(err: &VMLogicError) -> Self {
         match err {
             VMLogicError::HostError(h) => {
-                VMError::FunctionCallError(FunctionCallError::HostError(h))
+                VMError::FunctionCallError(FunctionCallError::HostError(h.clone()))
             }
-            VMLogicError::ExternalError(s) => VMError::ExternalError(s),
-            VMLogicError::InconsistentStateError(e) => VMError::InconsistentStateError(e),
+            VMLogicError::ExternalError(s) => VMError::ExternalError(s.clone()),
+            VMLogicError::InconsistentStateError(e) => VMError::InconsistentStateError(e.clone()),
         }
     }
 }
@@ -298,6 +294,7 @@ impl fmt::Display for PrepareError {
             StackHeightInstrumentation => write!(f, "Stack instrumentation failed."),
             Instantiate => write!(f, "Error happened during instantiation."),
             Memory => write!(f, "Error creating memory."),
+            #[cfg(feature = "protocol_feature_limit_contract_functions_number")]
             TooManyFunctions => write!(f, "Too many functions in contract."),
         }
     }
@@ -379,6 +376,7 @@ impl fmt::Display for VMError {
 impl std::fmt::Display for InconsistentStateError {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> Result<(), std::fmt::Error> {
         match self {
+            InconsistentStateError::StorageError(err) => write!(f, "Storage error: {:?}", err),
             InconsistentStateError::IntegerOverflow => write!(
                 f,
                 "Math operation with a value from the state resulted in a integer overflow.",
@@ -430,59 +428,27 @@ impl std::fmt::Display for HostError {
     }
 }
 
-/// Type-erased error used to shuttle some concrete error coming from `External`
-/// through vm-logic.
-///
-/// The caller is supposed to downcast this to a concrete error type they should
-/// know. This would be just `Box<dyn Any + Eq>` if the latter actually worked.
-pub struct AnyError {
-    any: Box<dyn AnyEq>,
-}
+pub mod hex_format {
+    use hex::{decode, encode};
 
-impl AnyError {
-    pub fn new<E: Any + Eq + Send + Sync + 'static>(err: E) -> AnyError {
-        AnyError { any: Box::new(err) }
-    }
-    pub fn downcast<E: Any + Eq + Send + Sync + 'static>(self) -> Result<E, ()> {
-        match self.any.into_any().downcast::<E>() {
-            Ok(it) => Ok(*it),
-            Err(_) => Err(()),
-        }
-    }
-}
+    use serde::de;
+    use serde::{Deserialize, Deserializer, Serializer};
 
-impl PartialEq for AnyError {
-    fn eq(&self, other: &Self) -> bool {
-        self.any.any_eq(&*other.any)
+    pub fn serialize<S, T>(data: T, serializer: S) -> Result<S::Ok, S::Error>
+    where
+        S: Serializer,
+        T: AsRef<[u8]>,
+    {
+        serializer.serialize_str(&encode(data))
     }
-}
 
-impl Eq for AnyError {}
-
-impl fmt::Debug for AnyError {
-    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        fmt::Debug::fmt(self.any.as_any(), f)
-    }
-}
-
-trait AnyEq: Any + Send + Sync {
-    fn any_eq(&self, rhs: &dyn AnyEq) -> bool;
-    fn as_any(&self) -> &dyn Any;
-    fn into_any(self: Box<Self>) -> Box<dyn Any>;
-}
-
-impl<T: Any + Eq + Sized + Send + Sync> AnyEq for T {
-    fn any_eq(&self, rhs: &dyn AnyEq) -> bool {
-        match rhs.as_any().downcast_ref::<Self>() {
-            Some(rhs) => self == rhs,
-            None => false,
-        }
-    }
-    fn as_any(&self) -> &dyn Any {
-        &*self
-    }
-    fn into_any(self: Box<Self>) -> Box<dyn Any> {
-        self
+    pub fn deserialize<'de, D, T>(deserializer: D) -> Result<T, D::Error>
+    where
+        D: Deserializer<'de>,
+        T: From<Vec<u8>>,
+    {
+        let s = String::deserialize(deserializer)?;
+        decode(&s).map_err(|err| de::Error::custom(err.to_string())).map(Into::into)
     }
 }
 
